import random
import numpy as np
try:
    from ctns.utility import retrive_to_test_quarantine, perform_test, compute_sd_reduction
except ImportError as e:
    from utility import retrive_to_test_quarantine, perform_test, compute_sd_reduction

def generate_family_edges(G):
    """
    Generate family edges. All edges between nodes of the same family are created
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network
        
    Return
    ------
    None

    """

    toAdd = []

    for node in G.vs:
        for edge in node["family_contacts"]:
            if G.vs[edge[0]]["agent_status"] != "D" \
            and G.vs[edge[1]]["agent_status"] != "D" \
            and not G.vs[edge[0]]["quarantine"] \
            and not G.vs[edge[1]]["quarantine"]:
                toAdd.append(edge)

    weights = np.random.randint(3, 8, len(toAdd))
    G.add_edges(toAdd)
    for (edge_index, i) in zip(G.get_eids(toAdd, directed = False), range(len(weights))):
      G.es[edge_index]["weight"] = weights[i]
      G.es[edge_index]["category"] = "family_contacts"  

def generate_occfreq_edges(G, edge_category, restriction_value):
    """
    Create edges from the node of type edge_category
    The number of edges is chosen according to the sociability of the node
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    edge_category: string
        Category of the edge attribute
    
    restriction_value: float
        How many edges are dropped in proportion to normal condition?
        
    Return
    ------
    None

    """

    toAdd = []

    for node in G.vs:
        possible_edges = node[edge_category].copy()
        random.shuffle(possible_edges)
        tmp = int(len(possible_edges) / 3)
        tmp2 = int(2 * len(possible_edges) / 3)
        if node["sociability"] == "low":
            n_edges = int(random.random() * tmp)
        if node["sociability"] == "medium":
            n_edges = tmp + int(random.random() * (tmp2 - tmp))
        if node["sociability"] == "high":
            n_edges = tmp2 + int(random.random() * (len(possible_edges) + 1 - tmp2))
        n_edges = int(n_edges * restriction_value)

        for i in range (0, n_edges):
            edge = possible_edges[i]
            if G.vs[edge[0]]["agent_status"] != "D" \
            and G.vs[edge[1]]["agent_status"] != "D" \
            and not G.vs[edge[0]]["quarantine"] \
            and not G.vs[edge[1]]["quarantine"] \
            and not G[edge[0], edge[1]]:
                toAdd.append(edge)

    weights = np.random.randint(1, 6, len(toAdd))
    G.add_edges(toAdd)
    for (edge_index, i) in zip(G.get_eids(toAdd, directed = False), range(len(weights))):
      G.es[edge_index]["weight"] = weights[i]
      G.es[edge_index]["category"]= edge_category  

def generate_random_edges(G, number_of_random_edges, restriction_value):
    """
    Create number_of_random_edges random edges in the contact network
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network
        
    number_of_random_edges : int
        Number of edges to create
    
    restriction_value: float
        How many edges are dropped in proportion to normal condition?
        
    Return
    ------
    None

    """

    toAdd = []
    number_of_random_edges = int(number_of_random_edges * restriction_value)
    edge_list = np.random.randint(0, len(list(G.vs)) - 1, 2 * number_of_random_edges)
    for i in range (0, 2 * number_of_random_edges, 2):
        source = edge_list[i]
        target = edge_list[i + 1]
        if not G[source, target] \
        and source != target \
        and G.vs[source]["agent_status"] != "D" \
        and G.vs[target]["agent_status"] != "D" \
        and not G.vs[source]["quarantine"] \
        and not G.vs[target]["quarantine"] \
        and not G[source, target]:
            toAdd.append((source, target))

    G.add_edges(toAdd)
    for edge_index in G.get_eids(toAdd, directed = False):
      G.es[edge_index]["weight"] = 1
      G.es[edge_index]["category"]= "random_contacts"  

def step_edges(G, restriction_value):
    """
    Removes old edges and creates new edges
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    restriction_value: float
        How many edges are dropped in proportion to normal condition?
        
    Return
    ------
    None

    """
    
    G.delete_edges(list(G.es))

    generate_family_edges(G)

    generate_occfreq_edges(G, "frequent_contacts", restriction_value)
    generate_occfreq_edges(G, "occasional_contacts", restriction_value)

    random_contact_total = len(list(G.vs)) + random.random() * (7 * len(list(G.vs)) - len(list(G.vs)))
    generate_random_edges(G, random_contact_total, restriction_value)

    # since edge generation produce a multigraph but a single edge has attributes as wanted,
    # all other edges are removed. This produce a simple graph
    # this operation is way much faster than checking befor adding new edge
    
    toRemove = list()
    for edge in G.es:
        if edge["weight"] == None:
            toRemove.append(edge)
    G.delete_edges(toRemove)  

def step_spread(G, incubation_days, infection_duration, transmission_rate, use_probabilities, alpha, gamma):
    """
    Make the infection spread across the network
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    incubation_days: int
        Average number of days where the patient is not infective

    infection_duration: int
        Average total duration of the disease
    
    transmission_rate: float
        Value of the transmission rate for the disease in the network

    use_probabilities: bool
        Enables probabilities of being infected estimation

    alpha: float
        Parameter to regulate probability of being infected contact decay. Domain = (0, 1). Lower values corresponds to higher probability decay

    gamma: float
        Parameter to regulate probability of being infected contact diffusion. Domain = (0, +inf). Higher values corresponds to stronger probability diffusion

    Return
    ------
    None
    
    """

    if use_probabilities:
        old_prob =  G.vs["prob_inf"]

    for node in G.vs:
        # update parameters if node is infected
        if node["infected"] == True:
            node["days_from_infection"] += 1
            if node["days_from_infection"] == infection_duration:
                # if infection is over, it will be dead of recovered
                node["agent_status"] = np.random.choice(["R", "D"], p = (1 - node["death_rate"], node["death_rate"]))
                node["infected"] = False
                node["symptoms"] = list()
                node["days_from_infection"] = 0
                node["needs_IC"] = False
                if node["agent_status"] == "D":
                    node["quarantine"] = 0
                    node["test_result"] = - 1
                    if use_probabilities:
                        node["prob_inf"] = 0
                
            # if it is still infective, spread the infection with his contacts
            if node["agent_status"] == "I":
                for contact in G.neighborhood(node)[1:]:
                    if G.vs[contact]["agent_status"] == "S":
                        prob = transmission_rate * G[node, contact] # access to the weight of the edge
                        if prob > 1:
                            prob = 1
                        # has the new node been infected?
                        G.vs[contact]["agent_status"] = np.random.choice(["S", "E"], p = (1 - prob, prob))
                        if G.vs[contact]["agent_status"] == "E":
                            G.vs[contact]["infected"] = True
                            G.vs[contact]["days_from_infection"] = 1

            # if the node become I, pick some symptoms
            if node["agent_status"] == "E" and node["days_from_infection"] == incubation_days:
                node["agent_status"] = "I"
                #if mild case
                case = random.uniform(0, 1)
                if case < 0.8:
                    if case < 0.05:
                        node["symptoms"].append("Loss of taste or smell")
                    if case < 0.2:
                        node["symptoms"].append("Fever")
                    if case < 0.2:
                        node["symptoms"].append("Cough")
                    if case < 0.2:
                        node["symptoms"].append("Tiredness")
                # if severe
                else:
                    if case < 0.99:
                        node["symptoms"].append("Fever")
                    if case < 0.7:
                        node["symptoms"].append("Tiredness")
                    if case < 0.6:
                        node["symptoms"].append("Cough")
                    if case < 0.3:
                        node["symptoms"].append("Dyspnea")
                
                if random.uniform(0, 1) < 0.02:
                    node["needs_IC"] = True

    # update prob of being infected
    if use_probabilities:
        nodes_contact_probs = [1] * len(list(G.vs))   
        
        for edge in G.es:
            weight = edge["weight"]
            source = edge.source
            target = edge.target
            nodes_contact_probs[source] *= 1 - (1 - 1 / (1 + alpha * 21)) * old_prob[target] * (1 - np.e**(- gamma * weight))
            nodes_contact_probs[target] *= 1 - (1 - 1 / (1 + alpha * 21)) * old_prob[source] * (1 - np.e**(- gamma * weight))

        for node in G.vs:
            if node["agent_status"] != "D" and not (node["test_result"] == 0 and node["agent_status"] == "R"):
                index = node.index
                node["prob_inf"] = 1 - (1 / (1 + alpha * 21)) * old_prob[index] - (1 - old_prob[index]) * nodes_contact_probs[index]
   
def step_test(G, nets, incubation_days, n_new_test, policy_test, contact_tracing_efficiency, quarantine_efficiency, use_probabilities, lambdaa):
    """
    Test some nodes of the network and put the in quarantine if needed
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    nets: list of ig.Graph()
        History of the network
    
    incubation_days: int
        Average number of days where the patient is not infective

    n_new_test: int
        Number of new avaiable tests

    policy_test: string
        Test strategy
        Can be ["Random, Degree Centrality, Betweenness Centrality, PBI"]
    
    contact_tracing_efficiency: float
        The percentage of contacts successfully traced

    quarantine_efficiency: float
        The percentage of contacts put under quarantine

    use_probabilities: bool
        Enables probabilities of being infected estimation

    lambdaa: float
        Parameter to regulate influence of contacts with a positive. Domain = (0, 1). Higher values corresponds to stronger probability diffusion

    Return
    ------
    new_positive_counter: int
        Number of new positive nodes found on this iteration

    """

    # create pool of nodes to test
    high_priority_test_pool = set()
    low_priority_test_pool = set()
    for node in G.vs:
        # update quarantine
        if node["quarantine"] > 0:
            node["quarantine"] -= 1
            # if node has been found positive and quarantine is over, re-test the node
            if node["quarantine"] == 0 and node["test_result"] == 1:
                high_priority_test_pool.add(node.index)
        # update test validity
        if node["test_validity"] > 0:
            node["test_validity"] -= 1
        # test also detect if the node is recovered (sierological test)
        # if node is not dead, if test validity is expired, if node is not a known recovered, add to low priority test pool
        if node["agent_status"] != "D" and node["test_validity"] <= 0 \
         and not (node["test_result"] == 0 and node["agent_status"] == "R"):
            low_priority_test_pool.add(node.index)

    low_priority_test_pool -= high_priority_test_pool
    found_positive = set()
    #cannot directly put nodes in a set cause node object are not supported
    high_priority_test_pool = [G.vs[i] for i in high_priority_test_pool]
    low_priority_test_pool = [G.vs[i] for i in low_priority_test_pool]
    for node in high_priority_test_pool:
        result = perform_test(node, incubation_days, use_probabilities)
        if result:
            found_positive.add(node.index)

    to_test = list()

    new_positive_counter = 0

    if n_new_test:
        if policy_test == "Random":
            to_test = random.sample(low_priority_test_pool, min(len(low_priority_test_pool), n_new_test))

        if policy_test == "Degree Centrality":
            low_priority_test_pool_index = [x.index for x in low_priority_test_pool]
            degree_results = G.strength(low_priority_test_pool_index, weights = "weight")
            to_test = retrive_to_test_quarantine(low_priority_test_pool, degree_results, n_new_test)

        if policy_test == "Betweenness Centrality":
            low_priority_test_pool_index = [x.index for x in low_priority_test_pool]
            betweenness_results = G.betweenness(low_priority_test_pool_index, 
                                                directed = False, weights = "weight",
                                                cutoff = None)
            to_test = retrive_to_test_quarantine(low_priority_test_pool, betweenness_results, n_new_test)

        if policy_test == "PBI":
            probs_infected = [i["prob_inf"] for i in low_priority_test_pool]
            to_test = retrive_to_test_quarantine(low_priority_test_pool, probs_infected, n_new_test)

    for node in to_test:
        result = perform_test(node, incubation_days, use_probabilities)
        if result:
            found_positive.add(node.index)
            new_positive_counter += 1
            
    # tracked will contain family contacts (quarantine 100%), 
    # possibly_tracked will contain other contacts, quarantine influenced by contact tracing efficiency
    ct_nets = list(nets) + [G.copy()]

    if len(found_positive) > 0:
        tracked = set()
        possibly_tracked = set()
        # trace contacts
        for net in ct_nets:
            for edge in net.es:
                if edge["category"] == "family_contacts" \
                and (edge.source in found_positive or edge.target in found_positive):
                    if edge.source in found_positive:
                        tracked.add(G.vs[edge.target].index)
                    else:
                        tracked.add(G.vs[edge.source].index)
                else:
                    if edge.source in found_positive:
                        possibly_tracked.add(G.vs[edge.target].index)
                    if edge.target in found_positive:
                        possibly_tracked.add(G.vs[edge.source].index)
        
        # set diff to remove double contacts
        possibly_tracked -= tracked
        if int(len(possibly_tracked) * contact_tracing_efficiency) > 0:
            possibly_tracked = random.sample(possibly_tracked, int(len(possibly_tracked) * contact_tracing_efficiency))
        else:
            possibly_tracked = list(possibly_tracked)
        tracked = list(tracked)
        if use_probabilities:
            # update prob of being infected of current contact 
            for node in tracked + possibly_tracked:
                for contact in G.neighborhood(node)[1:]:
                    contact_node = G.vs[contact]
                    if contact_node["agent_status"] != "D" and not (contact_node["test_result"] == 0 and contact_node["agent_status"] == "R"):
                        current_contact_weight = G[node, contact]
                        contact_node["prob_inf"] = contact_node["prob_inf"] \
                                                    + lambdaa * np.e**(-(1 / current_contact_weight)) * (1 - contact_node["prob_inf"])

            
            # update prob of being infected of past tracked contact 
                for net_index in range(1, len(nets) + 1):
                    net = nets[- net_index]
                    for contact in net.neighborhood(node)[1:]:
                        contact_node = G.vs[contact]
                        if contact_node["agent_status"] != "D" and not (contact_node["test_result"] == 0 and contact_node["agent_status"] == "R"):
                            if contact in tracked + possibly_tracked:
                                current_contact_weight = net[node, contact]
                                # use net_index instead of net_index + 1 since net index is already +1 from line above 
                                contact_node["prob_inf"] = contact_node["prob_inf"] \
                                                            + lambdaa * np.e**(- (net_index) * (1 / current_contact_weight)) * (1 - contact_node["prob_inf"])
        
        possibly_quarantine = list()

        if quarantine_efficiency == 1:
            possibly_quarantine = list(possibly_tracked)
        elif quarantine_efficiency == 0 or possibly_tracked == list():
            possibly_quarantine = list()
        else:
            if policy_test == "Random":
                possibly_quarantine = random.sample(possibly_tracked, int(len(possibly_tracked) * quarantine_efficiency))
            if policy_test == "Degree Centrality":
                values = G.strength(possibly_tracked, weights = "weight")
                possibly_quarantine = retrive_to_test_quarantine(possibly_tracked, values, int(len(possibly_tracked) * quarantine_efficiency))
            if policy_test == "Betweenness Centrality":
                values = G.betweenness(possibly_tracked, directed = False, weights = "weight", cutoff = None)
                possibly_quarantine = retrive_to_test_quarantine(possibly_tracked, values, int(len(possibly_tracked) * quarantine_efficiency))
            if policy_test == "PBI":
                values = [G.vs[x]["prob_inf"] for x in possibly_tracked]
                possibly_quarantine = retrive_to_test_quarantine(possibly_tracked, values, int(len(possibly_tracked) * quarantine_efficiency))

        to_quarantine = [G.vs[i] for i in tracked + possibly_quarantine]
        
        # put them in quarantine
        for node in to_quarantine:
            node["quarantine"] = 14

    return new_positive_counter

def step(G, step_index, incubation_days, infection_duration, transmission_rate,
         initial_day_restriction, restriction_duration, social_distance_strictness,
         restriction_decreasing, nets, n_test, policy_test, contact_tracing_efficiency,
         quarantine_efficiency, use_probabilities, alpha, gamma, lambdaa):
    """
    Advance the simulation of one step
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network
        
    step_index : int 
        index of the step

    incubation_days: int
        Average number of days where the patient is not infective

    infection_duration: int
        Average total duration of the disease
    
    transmission_rate: float
        Value of the transmission rate for the disease in the network
    
    initial_day_restriction: int
        Day index from when social distancing measures are applied

    restriction_duration: int
        How many days the social distancing last. Use -1 to make the restriction last till the end of the simulation

    social_distance_strictness: int
        How strict from 0 to 100 the social distancing measures are. 
        Represent the percentage of contact that are dropped in the network
        Note that family contacts are not included in this reduction

    restriction_decreasing: bool
        If the social distancing will decrease the strictness during the restriction_duration period

    nets: list of ig.Graph()
        History of the network

    n_test: int
        Number of avaiable tests

    policy_test: string
        Test strategy
    
    contact_tracing_efficiency: float
        The percentage of contacts successfully traced

    quarantine_efficiency: float
        The percentage of contacts put under quarantine

    use_probabilities: bool
        Enables probabilities of being infected estimation

    alpha: float
        Parameter to regulate probability of being infected contact decay. Domain = (0, 1). Lower values corresponds to higher probability decay

    gamma: float
        Parameter to regulate probability of being infected contact diffusion. Domain = (0, +inf). Higher values corresponds to stronger probability diffusion

    lambdaa: float
        Parameter to regulate influence of contacts with a positive. Domain = (0, 1). Higher values corresponds to stronger probability diffusion

    Return
    ------
    G: ig.Graph()
        The contact network

    new_positive_counter: int
        Number of new positive nodes found on this iteration

        
    """

    # generate new edges
    if not restriction_duration:
        if step_index >= initial_day_restriction:
            step_edges(G, 1 - (social_distance_strictness / 100))
        else:
            step_edges(G, 1)
    else:
        if step_index >= initial_day_restriction and step_index < initial_day_restriction + restriction_duration:
            if restriction_decreasing:
                social_distance_strictness = compute_sd_reduction(step_index, initial_day_restriction, restriction_duration, social_distance_strictness)
                step_edges(G, 1 - (social_distance_strictness / 100))
            else:
                step_edges(G, 1 - (social_distance_strictness / 100))
        else:
            step_edges(G, 1)
    
    # spread infection
    step_spread(G, incubation_days, infection_duration, transmission_rate, use_probabilities, alpha, gamma)
          
    # make some test on nodes
    new_positive_counter = step_test(G, nets, incubation_days, n_test, policy_test, contact_tracing_efficiency, quarantine_efficiency, use_probabilities, lambdaa)

    return G, new_positive_counter


import numpy as np
from pathlib import Path
from collections import deque, Counter
import sys, random, time, pickle
try:
    from ctns.generator import generate_network, init_infection, compute_TR
    from ctns.steps import step
    from ctns.utility import update_dump_report
except ImportError as e:
    from generator import generate_network, init_infection, compute_TR
    from steps import step
    from utility import update_dump_report

def run_simulation(n_of_families = 250,
    use_steps = True,
    number_of_steps = 150,
    incubation_days = 5,
    infection_duration = 21,
    initial_day_restriction = 50,
    restriction_duration = 21,
    social_distance_strictness = 2,
    restriction_decreasing = True,
    n_initial_infected_nodes = 3,
    R_0 = 2.9,
    n_test = 3,
    policy_test = "Random",
    contact_tracing_efficiency = 0.8,
    contact_tracing_duration = 14,
    quarantine_efficiency = 0.4,
    use_fixed_seed = True,
    seed = 42,
    use_probabilities = False,
    alpha = 0.5,
    gamma = 0.003,
    lambdaa = 0.02,
    dump_type = None,
    path = None):
    """
    Execute the simulation and dump/return resulting networks

    Parameters
    ----------
    n_of_families: int
        Number of families in the network
        
    use_steps : bool 
        Use a fixed number of steps or keep the simulation going untill the spreading is not over

    number_of_steps : int
        Number of simulation step to perform

    incubation_days: int
        Number of days where the patient is not infective

    infection_duration: int
        Total duration of the disease per patient

    initial_day_restriction: int
        Day index from when social distancing measures are applied

    restriction_duration: int
        How many days the social distancing last. Use 0 to make the restriction last till the end of the simulation

    social_distance_strictness: int
        How strict from 0 to 100 the social distancing measures are. 
        Represent the percentage of contact that are dropped in the network
        Note that family contacts are not included in this reduction
        Use 0 if you don't want to use restrictions

    restriction_decreasing: bool
        If the social distancing will decrease the strictness during the restriction_duration period

    n_initial_infected_nodes: int
        Number of nodes that are initially infected

    R_0: float
        The R0 facotr of the disease

    n_test: int
        Number of avaiable tests

    policy_test: string
        Strategy with which test are made. Can be Random, Degree Centrality, Betweenness Centrality or PBI

    contact_tracing_efficiency: float
        The percentage of contacts successfully traced back in the past contact_tracing_duration days

    contact_tracing_duration: int
        How many days the contacts of a node are tracked back

    quarantine_efficiency: float
        The percentage of contacts put under quarantine

    use_fixed_seed: bool
        Id use or not a fixed random seed

    seed: int
        The random seed value

    use_probabilities: bool
        Enables probabilities of being infected estimation

    alpha: float
        Parameter to regulate probability of being infected contact decay. Domain = (0, 1). Lower values corresponds to higher probability decay

    gamma: float
        Parameter to regulate probability of being infected contact diffusion. Domain = (0, +inf). Higher values corresponds to stronger probability diffusion

    lambdaa: float
        Parameter to regulate influence of contacts with a positive. Domain = (0, 1). Higher values corresponds to stronger probability diffusion

    dump_type: string
        Can be either ["full", "light"]. In the first case, the full simulation (all nets structure) is dumped.
        Otherwise, only a report about node status is saved.
        NB, full method will use significally more RAM than light; also the dump will have much bigger size.
        The dumped file will have the following structure:
        - a dict containig simulation parameters and a list of ig.Graph() if dump_type is full
        - a dict[class] where class can be [S, E, I, R, D, quarantined, positive, tested, total] and value is a list of the corresponding attribute value on day i

    path: string
        The path to the file/folder where the networks will be saved

    Return
    ------
    None

    """

    # generate new edges
    if use_fixed_seed:
        np.random.seed(seed = seed)
        random.seed(seed)
    else:
        np.random.seed(int(time.time()))
        random.seed(time.time())
    
    # check values
    if n_of_families < 10:
        print("Invalid number of families. Use at least 10 families")
        sys.exit()
    if use_steps:
        if number_of_steps < 0:
            print("Invalid number of steps")
            sys.exit()
    if infection_duration < 0:
        print("Invalid infection duration")
        sys.exit()
    if incubation_days < 0 or incubation_days >= infection_duration:
        print("Invalid incubation duration")
        sys.exit()
    if initial_day_restriction < 0 :
        print("Invalid initial day social distancing")
        sys.exit()
    if social_distance_strictness < 0 or social_distance_strictness > 100:
        print("Invalid social distancing value")
        sys.exit()
    if n_initial_infected_nodes < 0 or n_initial_infected_nodes > n_of_families:
        print("Invalid number of initial infected nodes")
        sys.exit()
    if R_0 < 0:
        print("Invalid value of R0")
        sys.exit()
    if n_test < 0:
        print("Invalid number of test per day")
        sys.exit()
    if policy_test not in ["", "Random", "Degree Centrality", "Betweenness Centrality", "PBI"]:
        print("Invalid test strategy")
        sys.exit()
    if contact_tracing_efficiency < 0 or contact_tracing_efficiency > 1:
        print("Invalid contact tracing efficiency")
        sys.exit()
    if quarantine_efficiency < 0 or quarantine_efficiency > 1:
        print("Invalid quarantine efficiency")
        sys.exit()
    if contact_tracing_duration < 0:
        print("Invalid contatc tracing duration")
        sys.exit()
    if restriction_duration < 0:
        print("Invalid restriction restriction_duration")
        sys.exit()
    if policy_test == "":
        n_test = 0
    if gamma < 0 :
        print("Value of gamma must greater than 0")
        sys.exit()
    if lambdaa < 0 or lambdaa > 1:
        print("Value of lambda must be 0 < lambda < 1")
        sys.exit()
    if dump_type != "full" and dump_type != "light":
        print("Invalid dump type")
        sys.exit()
    if path == None:
        print("Invalid path")
        sys.exit()
    if not use_probabilities and policy_test == "PBI":
        print("Cannot use PBI if probability of being infected is not enabled")
        sys.exit()

    # making parameters consistent
    if social_distance_strictness == 0:
        restriction_decreasing = False
        initial_day_restriction = 0
        restriction_duration = 0

    if n_test == 0:
        policy_test = None
        contact_tracing_efficiency = 0
        contact_tracing_duration = 0

    if contact_tracing_duration == 0 or contact_tracing_efficiency == 0:
        contact_tracing_duration = 0
        contact_tracing_efficiency = 0

    if not use_probabilities:
        gamma = 0
        lambdaa = 0

    config = locals()
    a = time.perf_counter()
    # init network
    G = generate_network(n_of_families, use_probabilities)
    transmission_rate = compute_TR(G, R_0, infection_duration, incubation_days)
    init_infection(G, n_initial_infected_nodes)

    nets = deque(maxlen = contact_tracing_duration)
    if dump_type == "full":
        to_dump = dict()
        to_dump["nets"] = list()
        to_dump["parameters"] = config
    if dump_type == "light":
        to_dump = dict()
        to_dump['S'] = list()
        to_dump['E'] = list()
        to_dump['I'] = list()
        to_dump['R'] = list()
        to_dump['D'] = list()
        to_dump['quarantined'] = list()
        to_dump['positive'] = list()
        to_dump['tested'] = list()
        to_dump['total'] = list()
        to_dump['new_positive_counter'] = list()
        to_dump['avg_prob_inf'] = list()
        to_dump['parameters'] = config

    if use_steps:
        for sim_index in range (0, number_of_steps):
            net, new_positive_counter = step(G, sim_index, incubation_days, infection_duration, transmission_rate,
                             initial_day_restriction, restriction_duration, social_distance_strictness, 
                             restriction_decreasing, nets, n_test, policy_test, contact_tracing_efficiency,
                             quarantine_efficiency, use_probabilities, alpha, gamma, lambdaa)
            nets.append(net.copy())
            if dump_type == "full":
                to_dump["nets"].append(net.copy())
            if dump_type == "light":
                to_dump = update_dump_report(to_dump, net, new_positive_counter, use_probabilities)
    else:
        exposed = n_initial_infected_nodes
        infected = 0
        sim_index = 0
        while((infected + exposed) != 0):
            net, new_positive_counter = step(G, sim_index, incubation_days, infection_duration, transmission_rate,
                             initial_day_restriction, restriction_duration, social_distance_strictness, 
                             restriction_decreasing, nets, n_test, policy_test, contact_tracing_efficiency,
                             quarantine_efficiency, use_probabilities, alpha, gamma, lambdaa)
            nets.append(net.copy())
            sim_index += 1

            if dump_type == "full":
                to_dump["nets"].append(net.copy())
            if dump_type == "light":
                to_dump = update_dump_report(to_dump, net, new_positive_counter, use_probabilities)
            
            network_report = Counter(net.vs["agent_status"])
            infected = network_report["I"]
            exposed = network_report["E"]
            if infected + exposed == 0:
                break

    with open(Path(path + ".pickle"), "wb") as f:
        pickle.dump(to_dump, f, protocol = pickle.DEFAULT_PROTOCOL)
    b = time.perf_counter()
    print("\n Simulation ended successfully \n")
    print("time elapsed " + str(b-a))

def main():

    dump_type = None
    path = None

    user_interaction = int(input("Press 0 to load the default values or 1 to manually input the configuration for the simulation: "))
    # get values from user
    if user_interaction:
        n_of_families = int(input("Please insert the number of families in the simulation: "))
        use_steps = int(input("Press 1 use a fixed number of steps or 0 to run the simulation untill the infection is over: "))
        if use_steps:
            number_of_steps = int(input("Please insert the number of the simulation step: "))
        incubation_days = int(input("Please insert the disease incubation duration: "))
        infection_duration = int(input("Please insert the disease duration: "))
        initial_day_restriction = int(input("Please insert the step index from which the social distance is applied: "))
        restriction_duration = int(input("Please insert the number of days which the social distance last. Insert 0 to make the restriction last for all the simulation: "))
        social_distance_strictness = int(input("Please insert a value between 0 and 4 to set the social distance strictness: "))
        restriction_decreasing = int(input("Press 1 to make the strictness of the social distance decrease during the simulation or 0 to keep it fixed: "))
        n_initial_infected_nodes = int(input("Please insert the number of initial infected individuals: "))
        R_0 = float(input("Please insert the value of R0: "))
        n_test = int(input("Please insert the number of available test per day: "))
        policy_test = input("Please insert strategy with which test are made. Can be Random, Degree Centrality, Betweenness Centrality or PBI (probability of being infected): ")
        contact_tracing_efficiency = float(input("Please insert a value between 0 and 1 to set the contact tracing efficiency: "))
        contact_tracing_duration = int(input("Please insert for how many days the contact tracing is computed: "))
        quarantine_efficiency = float(input("Please insert a value between 0 and 1 to set the quarantine efficiency: "))
        use_fixed_seed = int(input("Press 1 use a fixed a random seed or 0 to pick a random seed: "))
        if use_fixed_seed:
            seed = int(input("Please insert the random seed: "))
        use_probabilities = int(input("Press 1 to enable estimation of probabilities of being infected, 0 otherwise: "))
        if use_probabilities:
            alpha = float(input("Please insert the value of alpha: "))
            gamma = float(input("Please insert the value of gamma: "))
            lambdaa = float(input("Please insert the value of lambdaa: "))
        dump_type = input("Please insert the dump type. Can be either full of light: ")
        path = input("Please insert the path with the file to dump. Please omit file type, that will be set automatically: ")

        run_simulation(n_of_families, use_steps, number_of_steps, incubation_days, infection_duration,
            initial_day_restriction, restriction_duration, social_distance_strictness, restriction_decreasing,
            n_initial_infected_nodes, R_0, n_test, policy_test, contact_tracing_efficiency, contact_tracing_duration,
            quarantine_efficiency, use_fixed_seed, seed, use_probabilities, alpha, gamma, lambdaa, dump_type, path)
    else:
        dump_type = input("Please insert the dump type. Can be either full of light: ")
        path = input("Please insert the path with the file to dump. Please omit file type, that will be set automatically: ")
        run_simulation(path = path, dump_type = dump_type)

if __name__ == "__main__":
    main()


import numpy as np
from collections import Counter

def fix_distribution_node_number(distribution, n_nodes):
    """
    Make the sum of the elements in the distribution be equal to the number of nodes
    
    Parameters
    ----------
    distribution: list of int
        Distribution of int

    n_nodes: int
        Number of nodes in The contact network
        
    Return
    ------
    distribution: list of int
        Distribution where the sum of elements is equal to n_nodes

    """

    refined_distribution = list()
    while True:
        refined_distribution.append(distribution.pop())
        if (np.sum(refined_distribution) > n_nodes):
            refined_distribution.pop()
            break
    refined_distribution.append(int(n_nodes - np.sum(refined_distribution)))
    return refined_distribution  

def reset_network(G):
    """
    Reset network status, removing all edges and setting all the infection 
    related attributes of each node to the default value
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network
        
    Return
    ------
    None

    """

    G.delete_edges(list(G.es))
    for node in G.vs:
        node["agent_status"] = 'S'
        node["infected"] = False
        node["days_from_infection"] = 0
        node["quarantine"] = 0
        node["test_validity"] = 0
        node["test_result"] = -1
        node["symptoms"] = list()
        if "prob_inf" in G.vs.attributes():
            node["prob_inf"] = 0.0

def update_dump_report(to_dump, net, new_positive_counter, use_probabilities):
    """
    Update the simulation dump in case light dump is selected
    
    Parameters
    ----------
    to_dump: dict
        Old dump doctionary

    net: ig.Graph()
        The contact network

    new_positive_counter: int
        Number of new positive nodes found on this iteration

    use_probabilities: bool
        Enables probabilities of being infected estimation

    Return
    ------
    to_dump: dict
        Updated dump doctionary

    """

    network_report = Counter(net.vs["agent_status"])
    tested = 0
    positive = 0
    quarantined = 0
    for node in net.vs:
        if node["test_result"] != -1:
            tested += 1
        if node["test_result"] == 1:
            positive += 1
        if node["quarantine"] != 0:
            quarantined += 1

    to_dump['S'].append(network_report['S'])
    to_dump['E'].append(network_report['E'])
    to_dump['I'].append(network_report['I'])
    to_dump['R'].append(network_report['R'])
    to_dump['D'].append(network_report['D'])
    to_dump['quarantined'].append(quarantined)
    to_dump['positive'].append(positive)
    to_dump['tested'].append(tested)
    to_dump['new_positive_counter'].append(new_positive_counter)
    if use_probabilities:
        to_dump['avg_prob_inf'].append(np.mean(net.vs['prob_inf']))
    to_dump['total'].append(sum(network_report.values()))

    return to_dump

def compute_sd_reduction(step_index, initial_day_restriction, restriction_duration, social_distance_strictness):
    """
    Calculate the decreased social distance strictness index
    
    Parameters
    ----------
        
    step_index : int 
        index of the step
    
    initial_day_restriction: int
        Day index from when social distancing measures are applied

    restriction_duration: int
        How many days the social distancing last. Use -1 to make the restriction last till the end of the simulation

    social_distance_strictness: int
        How strict from 0 to 4 the social distancing measures are. 
        Represent the portion of contact that are dropped in the network (0, 25%, 50%, 75%, 100%)
        Note that family contacts are not included in this reduction

    Return
    ------
    social_distance_strictness: int
        Updated value for social_distance_strictness
        
    """
    
    default_days = restriction_duration // social_distance_strictness
    spare_days = restriction_duration - default_days * social_distance_strictness

    strictness_sequence = list()

    for i in range(social_distance_strictness):
        updated_days = default_days
        if spare_days > 0:
            updated_days += 1
            spare_days -= 1
        strictness_sequence.append(updated_days)


    for i in range(1, len(strictness_sequence)):
      strictness_sequence[i] += strictness_sequence[i - 1]

    social_distance_reduction = None

    for i in range(len(strictness_sequence)):
      if step_index - initial_day_restriction < strictness_sequence[i]:
        social_distance_reduction = i
        break

    return social_distance_strictness - social_distance_reduction

def perform_test(node, incubation_days, use_probabilities):
    """
    Perform tampon and syerological test on the node
    
    Parameters
    ----------
        
    node: igraph.Vertex
        Node to test

    incubation_days: int
        Average number of days where the patient is not infective

    use_probabilities: bool
        Enables probabilities of being infected estimation

    Return
    ------
    positive: bool
        If the node tested is found positive
        
    """

    if node["infected"]:
        node["test_result"] = 1
        node["quarantine"] = 14
        node["test_validity"] = 14
        if use_probabilities:
            node["prob_inf"] = 1
        return True
    else:
        node["test_result"] = 0
        node["test_validity"] = incubation_days
        if use_probabilities:
            node["prob_inf"] = 0
        return False

    return None

def retrive_to_test_quarantine(nodes, values, n_new_test, reverse = True):
    """
    Test some nodes of the network and put the in quarantine if needed
    
    Parameters
    ----------

    nodes: list
        List of all nodes that needs a test

    values: list
        List of values used to sort the previous list

    n_new_test: int
        Number of new avaiable tests

    reverse: bool
        Sort using descending order

    Return
    ------
    selected: list
        List of nodes that will be tested

    """

    selected = list()

    zipped_lists = zip(values, nodes)
    sorted_pairs = sorted(zipped_lists, reverse = reverse)
    tuples = zip(*sorted_pairs)
    sorted_values, sorted_nodes = [list(tuple) for tuple in  tuples]
    selected = sorted_nodes[:min(n_new_test, len(nodes))]

    return selected


from setuptools import setup, find_packages
import os

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open('README.md') as readme_file:
    README = readme_file.read()

ver = os.environ['CI_COMMIT_TAG']

setup(
  name = 'ctns',
  packages = find_packages(),
  scripts = ['bin/ctns', 'bin/ctns.bat','bin/ctns_run.py'],
  description = 'CTNS, Contact Tracing Network Simulator: a tool to simulate contact tracing in a population where a disease is spreading',
  long_description_content_type = "text/markdown",
  long_description = README,
  license = 'GNU GPLv3',
  author = 'Matteo Mistri, Diego Miglio',
  author_email = 'matteo.mistri1996@gmail.com',
  install_requires = required,
  version = ver,
  url = "https://gitlab.com/mistrello96/ctns",
  download_url='https://pypi.org/project/ctns/',
)